#include <htc.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include  <ctype.h>
#include "init.h"
#include "uart.h"
#include "lcd.h"

// PIC has 4MHz clock
#define _XTAL_FREQ 4000000

//****************************
// Special methods - utilities
//****************************

/*
 * A copy of standard strsplit C method that does not skip leading delimeters.
 */
static char * sp;
char * strsplit(register char * s1, register const char * s2) {
  if (!s1)
    s1 = sp;
  if (!s1)
    return NULL;
  // s1 += strspn(s1, s2); /* skip leading delimiters */
  if (!*s1)
    return sp = NULL;
  sp = s1 + strcspn(s1, s2);
  if (*sp)
    *sp++ = 0;
  else
    sp = 0;
  return s1;
}

/*
 * Converts one character to integer. If character is not digit, 0 is returned.
 */
int ctoi(register char c) {
  if (isdigit(c)) {
    return c-'0';
  }
  return 0;
}

//****************************
// Variables
//****************************

// Time in seconds, when site is changed on LCD
#define SITE_SEC 5

boolean fix=false; // GPS valid fix
int year;
int month;
int day;
int hour;
int minute;
int second;
int latDeg; // Latitude degrees
float latMin; // Latitude minutes
boolean east; // E/W indicator
int longDeg; // Longitude degrees
float longMin; // Longitude minutes
boolean north; // N/S indicator
char satellites; // Number of found satellites
float altitude; // MSL altitude (geoid)
float speed; // Speed over ground in km/h
float course; // Course over ground

//****************************
// Rendering
//****************************

/*
 * Render site 1.
 */
void render1(void) {
  int h;

  // Convert from UTC to CET - add one hour
  h=(hour+1)%24;

  lcd_returnHome();
  printf("20%.2d-%.2d-%.2d      ", year, month, day);
  lcd_nextLine();
  printf("%.2d:%.2d:%.2d %6dm", h, minute, second, (int)altitude);
}

/*
 * Render site 2.
 */
void render2(void) {
  int longMinDecimal;
  int latMinDecimal;
  
  // Get one decimal number for minutes
  longMinDecimal=(int)((longMin-(int)longMin)*10);
  latMinDecimal=(int)((latMin-(int)latMin)*10);

  lcd_returnHome();
  printf("%3d*%.2d.%d'%c %3dkh", longDeg, (int)longMin, longMinDecimal,
      east?'E':'W', (int)speed);
  lcd_nextLine();
  printf("%3d*%.2d.%d'%c %3d* ", latDeg, (int)latMin, latMinDecimal,
      north?'N':'S', (int)course);
}

/*
 * Displays data on LCD.
 */
boolean site2=false;
unsigned int renderCount=0;
void render(void) {
  // Display 'Searching satellites' when fix is invalid
  if (!fix) {
    lcd_returnHome();
    printf("Searching       ");
    lcd_nextLine();
    printf("satellites    %2d", satellites);
    return;
  }

  if (site2) {
    render2();

  } else {
    render1();
  }

  if (renderCount%SITE_SEC==0) {
    // Time to change site
    site2=!site2;
  }
  renderCount++;
}

//****************************
// Parsing
//****************************

char * pch;

/*
 * Parses the given GPGGA sentence.
 */
void parseGpgga(char * gpgga) {
  pch=strsplit(gpgga, ",");
  if (pch==NULL) return;
  // $GPGGA, skip

  pch=strsplit(NULL, ",");
  if (pch==NULL) return;
  // UTC time
  hour=ctoi(pch[0])*10+ctoi(pch[1]);
  minute=ctoi(pch[2])*10+ctoi(pch[3]);
  second=ctoi(pch[4])*10+ctoi(pch[5]);

  pch=strsplit(NULL, ",");
  if (pch==NULL) return;
  // Latitude
  latDeg=ctoi(pch[0])*10+ctoi(pch[1]);
  latMin=atof(&pch[2]);

  pch=strsplit(NULL, ",");
  if (pch==NULL) return;
  // N/S indicator
  north=(pch[0]=='N');

  pch=strsplit(NULL, ",");
  if (pch==NULL) return;
  // Longitude
  longDeg=ctoi(pch[0])*100+ctoi(pch[1])*10+ctoi(pch[2]);
  longMin=atof(&pch[3]);

  pch=strsplit(NULL, ",");
  if (pch==NULL) return;
  // E/W indicator
  east=(pch[0]=='E');

  pch=strsplit(NULL, ",");
  if (pch==NULL) return;
  // Position fix indicator
  fix=(pch[0]!='0');

  pch=strsplit(NULL, ",");
  if (pch==NULL) return;
  // Satellites
  satellites=atoi(pch);

  pch=strsplit(NULL, ",");
  if (pch==NULL) return;
  // HDOP, skip

  pch=strsplit(NULL, ",");
  if (pch==NULL) return;
  // MSL altitude (geoid)
  altitude=atof(pch);
}

/*
 * Parses the given GPRMC sentence.
 */
void parseGprmc(char * gprmc) {
  pch=strsplit(gprmc, ",");
  if (pch==NULL) return;
  // $GPRMC, skip

  pch=strsplit(NULL, ",");
  if (pch==NULL) return;
  // UTC time, skip - got it from GPGGA

  pch=strsplit(NULL, ",");
  if (pch==NULL) return;
  // Status, skip

  pch=strsplit(NULL, ",");
  if (pch==NULL) return;
  // Latitude, skip - got it from GPGGA

  pch=strsplit(NULL, ",");
  if (pch==NULL) return;
  // N/S indicator, skip - got it from GPGGA

  pch=strsplit(NULL, ",");
  if (pch==NULL) return;
  // Longitude, skip - got it from GPGGA

  pch=strsplit(NULL, ",");
  if (pch==NULL) return;
  // E/W indicator, skip - got it from GPGGA

  pch=strsplit(NULL, ",");
  if (pch==NULL) return;
  // Speed over ground
  speed=atof(pch);
  // Convert from knots to km/h
  speed*=1.852;

  pch=strsplit(NULL, ",");
  if (pch==NULL) return;
  // Course over ground
  course=atof(pch);

  pch=strsplit(NULL, ",");
  if (pch==NULL) return;
  // Date
  day=ctoi(pch[0])*10+ctoi(pch[1]);
  month=ctoi(pch[2])*10+ctoi(pch[3]);
  year=ctoi(pch[4])*10+ctoi(pch[5]);
}

/*
 * Tries to parse given NMEA sentence. Returns true if sentence is GPGGA.
 */
boolean parse(char * nmea) {
  if (strncmp(nmea, "$GPGGA", 6)==0) {
    // Got GPGGA, parse fully
    parseGpgga(nmea);
    return true;

  } else if (strncmp(nmea, "$GPRMC", 6)==0) {
    // Got GPRMC, parse fully
    parseGprmc(nmea);
  }
  return false;
}

//****************************
// Main
//****************************

void main(void) {
  int i;
  boolean gpgga;
  // NMEA sentece from GPS
  unsigned char * nmea;

  // Init PIC
  init();

  // Set up UART
  uart_setup();

  // Set up LCD
  lcd_setup();

  // Display author
  printf("  Tine Lesjak   ");
  lcd_nextLine();
  printf("    (C) 2009    ");

  // Wait for ~3 secs
  for (i=0; i<15; i++) {
    __delay_ms(197);
  }

  // Main loop
  while (true) {
    // Get line from GPS
    nmea=uart_getLine();

    // Parse line
    gpgga=parse(nmea);

    // If line was GPGGA (every 1 second), display data
    if (gpgga) {
      render();
    }
  }
}
